import muse.data_interface as di
from muse.func_desc import FUNC_DESC

def _parse_products(stms: list):
    assets = list()
    ports = list()
    for index, s in enumerate(stms):
        if '函数调用:资产' in s or '函数调用:产品' in s:
            tags = list()
            indicators = list()
            start = index
            param_index = index + 1
            if param_index < len(stms):
                if '函数参数1' in stms[param_index]:
                    param_index = param_index + 1
                    if param_index < len(stms):
                        for m in stms[param_index:]:
                            if '字符串字面值:' in m:
                                tags.append(m.split(':')[1].strip())
                                param_index = param_index + 1
                            elif '数组' in m:
                                param_index = param_index + 1
                                continue
                            else:
                                break
            if param_index < len(stms):
                if '函数参数2' in stms[param_index]:
                    param_index = param_index + 1
                    if param_index < len(stms):
                        for m in stms[param_index:]:
                            if '字符串字面值:' in m:
                                indicators.append(m.split(':')[1].strip())
                                param_index = param_index + 1
                            elif '数组' in m:
                                param_index = param_index + 1
                                continue
                            else:
                                break
            if '函数调用:资产' in s:
                assets.append({'标签': tags, '指标': indicators, '位置': (start, param_index)})
            else:
                ports.append({'标签': tags, '指标': indicators, '位置': (start, param_index)})
    return assets, ports


def parse_info(stms: list):
    last_s = stms[-1]

    tags = list()
    indicators = list()
    funcs = list()
    variables = list()
    for s in stms:
        if '函数调用:' in s:
            funcs.append(s.split(':')[1])
        if '变量:' in s:
            variables.append(s.split(':')[1])
    colors = dict()
    funcs = list(set(funcs))
    colors['函数'] = funcs
    variables = list(set(variables))
    colors['变量'] = variables
    assets, ports = _parse_products(stms)
    for a in assets:
        tags = tags + a['标签']
        indicators = indicators + a['指标']
    for p in ports:
        tags = tags + p['标签']
        indicators = indicators + p['指标']
    tags = list(set(tags))
    indicators = list(set(indicators))
    colors['标签'] = tags
    colors['指标'] = indicators

    prompts = {}

    last_index = len(stms) - 1
    # print(last_s)
    # 情况一： AST最后一个元素为[变量]：提示之前定义过的变量或者固有的内部函数
    # 例子：回购资产 = 回TAB
    # 返回前端： ‘prompts’: {‘变量’: [所有之前定义的变量列表], ‘函数’: [所有支持的函数列表]}
    if '变量:' in last_s or '数组第' in last_s:
        if len(assets) > 0 and assets[-1]['位置'][1] == last_index:
                v = last_s.split(':')[1].strip()
                # 继续向前找看是那个函数参数
                for s in reversed(stms):
                    if '函数参数1' in s:
                        asset_tags = di.ASSET_INDS
                        if v != '':
                            asset_tags = list()
                            for t in di.ASSET_INDS:
                                if v in t:
                                    asset_tags.append(t)
                        prompts = {'资产标签': asset_tags, '原文': v}
                        break
                    if '函数参数2' in s:
                        asset_inds = di.HLD_INDS
                        if v != '':
                            asset_inds = list()
                            for t in di.HLD_INDS:
                                if v in t:
                                    asset_inds.append(t)
                        prompts = {'资产指标': asset_inds, '原文': v}
                        break

        elif len(ports) > 0 and ports[-1]['位置'][1] == last_index:
                v = last_s.split(':')[1].strip()
                # 继续向前找看是那个函数参数
                for s in reversed(stms):
                    if '函数参数1' in s:
                        port_tags = di.PORT_TAGS
                        if v != '':
                            port_tags = list()
                            for t in di.PORT_TAGS:
                                if v in t:
                                    port_tags.append(t)
                        prompts = {'产品标签': port_tags, '原文': v}
                        break
                    if '函数参数2' in s:
                        port_inds = di.PORT_INDS
                        if v != '':
                            port_inds = list()
                            for t in di.PORT_INDS:
                                if v in t:
                                    port_inds.append(t)
                        prompts = {'产品指标': port_inds, '原文': v}
                        break
        elif '变量:' in last_s:
            variable = last_s.split(':')[1].strip()
            if '指标部分:' in stms[-2]:
                inds = list()
                for ind in indicators + di.BASIC_COLS:
                    if variable in ind:
                        inds.append(ind)
                prompts = {'指标': inds, '原文': variable}
            else:
                f_list = []
                for f in di.VALID_FUNCS:
                    if variable in f:
                        f_list.append(f)
                var_list = []
                for v in variables:
                    if variable in v:
                        var_list.append(v)
                if variable in var_list:
                    var_list.remove(variable)
                if variable in colors['变量']:
                    colors['变量'].remove(variable)
                prompts = {'变量': var_list, '函数': f_list, '原文': variable}
    # AST最后一个是[函数参数1]：向前找到函数的名称，提示函数的调用方法
    # 例子：回购资产 = agg(TAB
    # 返回前端：’prompts’: {‘函数’: {函数名: agg, 函数描述: 函数描述}}
    elif '函数参数1' in last_s:
        for s in reversed(stms):
            if '函数调用:' in s:
                func_name = s.split(':')[1].strip()
                if func_name in FUNC_DESC:
                    func_desc = FUNC_DESC[func_name]
                else:
                    func_desc = func_name
                prompts = {'函数': {'函数名': func_name, '函数描述': func_desc}, '原文': func_name}
                break
    elif '指标部分:' in last_s:
        prompts = {'指标': indicators + di.BASIC_COLS, '原文': ''}

    # print(assets)
    # print(ports)
    # print(last_index)
    return colors, prompts


# 四则运算解析
# Dataframe之间的广播类和向量计算
# 先merge相关dataframe再进行计算
from muse.scanner import Op
import polars as pl
import logging


def handle_in_op(exp1, exp2):
    logging.debug('调用in处理函数 {} in {}; 调用结果: {}'.format(exp1, exp2, 5))
    return 5


def handle_binary_op(op: Op, exp1, exp2):
    # 准备默认值
    if isinstance(exp1, tuple):
        exp1_var, exp1 = exp1
    else:
        exp1_var, exp1 = None, exp1

    if isinstance(exp2, tuple):
        exp2_var, exp2 = exp2
    else:
        exp2_var, exp2 = None, exp2

    if isinstance(exp1, pl.DataFrame) and exp1.is_empty():
        return pl.DataFrame()
    if isinstance(exp2, pl.DataFrame) and exp2.is_empty():
        return pl.DataFrame()

        # DataFrame与标量之间的操作
    if exp1_var is not None and exp2_var is None:
        exp1 = perform_operation_on_df_and_scalar(op, exp1, exp1_var, exp2)
        result = exp1

        # 标量与DataFrame之间的操作
    elif exp1_var is None and exp2_var is not None:
        exp2 = perform_operation_on_scalar_and_df(op, exp1, exp2, exp2_var)
        result = exp2

        # 两个都是DataFrame
    elif exp1_var is not None and exp2_var is not None:
        result = perform_operation_on_dfs(op, exp1, exp2, exp1_var, exp2_var)

        # 两个都是标量
    else:
        result = perform_operation_on_scalars(op, exp1, exp2)

    logging.debug('调用二元运算: {} {} {}; 调用结果: {}'.format(exp1, op.name, exp2, result))
    return result


def perform_operation_on_df_and_scalar(op, exp1, exp1_var, exp2):
    if op == Op.Plus:
        return exp1.with_columns((pl.col(exp1_var) + exp2).alias(exp1_var))
    elif op == Op.Minus:
        if exp1.select(exp1_var).to_series().dtype == pl.String and isinstance(exp2, str):
            return exp1.with_columns(
                (pl.col(exp1_var).str.to_datetime("%Y-%m-%d") - pl.Series([exp2]).str.to_datetime("%Y-%m-%d"))
                .dt.total_days().alias(exp1_var))
        else:
            return exp1.with_columns((pl.col(exp1_var) - exp2).alias(exp1_var))
    elif op == Op.Multiply:
        return exp1.with_columns((pl.col(exp1_var) * exp2).alias(exp1_var))
    elif op == Op.Divide:
        return exp1.with_columns((pl.col(exp1_var) / exp2).alias(exp1_var))


def perform_operation_on_scalar_and_df(op, exp1, exp2, exp2_var):
    if op == Op.Plus:
        return exp2.with_columns((exp1 + pl.col(exp2_var)).alias(exp2_var))
    elif op == Op.Minus:
        if exp2.select(exp2_var).to_series().dtype == pl.String and isinstance(exp1, str):
            return exp2.with_columns(
                pl.Series([exp1]).str.to_datetime("%Y-%m-%d") - (pl.col(exp2_var).str.to_datetime("%Y-%m-%d"))
                .dt.total_days().alias(exp2_var))
        else:
            return exp2.with_columns((exp1 - pl.col(exp2_var)).alias(exp2_var))
    elif op == Op.Multiply:
        return exp2.with_columns((exp1 * pl.col(exp2_var)).alias(exp2_var))
    elif op == Op.Divide:
        return exp2.with_columns((exp1 / pl.col(exp2_var)).alias(exp2_var))


def perform_operation_on_dfs(op, exp1, exp2, exp1_var, exp2_var):
    # 打印并清理列重命名
    common_cols = list(set(exp1.columns).intersection(exp2.columns))
    if len(common_cols) == len(exp1.columns) and len(common_cols) == len(exp2.columns):
        df = exp1
    else:
        df = exp1.join(exp2, on=common_cols, how='left')

    if op == Op.Plus:
        result = df.with_columns((pl.col(exp1_var) + pl.col(exp2_var)).alias(exp1_var + '+' + exp2_var))
    elif op == Op.Minus:
        if df.select(exp1_var).to_series().dtype == pl.String and df.select(exp2_var).to_series().dtype == pl.String:
            result = df.with_columns(
                (pl.col(exp1_var).str.to_datetime("%Y-%m-%d") - pl.col(exp2_var).str.to_datetime("%Y-%m-%d"))
                .dt.total_days().cast(pl.Int64).alias(exp1_var + '-' + exp2_var))
        elif df.select(exp1_var).to_series().dtype == pl.UInt16 and df.select(exp2_var).to_series().dtype == pl.UInt16:
            result = df.with_columns(
                (pl.col(exp1_var).cast(pl.Int64) - pl.col(exp2_var).cast(pl.Int64)).cast(pl.Int64).alias(f"{exp1_var}-{exp2_var}")
            )
        else:
            result = df.with_columns((pl.col(exp1_var) - pl.col(exp2_var)).alias(exp1_var + '-' + exp2_var))
    elif op == Op.Multiply:
        result = df.with_columns((pl.col(exp1_var) * pl.col(exp2_var)).alias(exp1_var + '*' + exp2_var))
    elif op == Op.Divide:
        result = df.with_columns((pl.col(exp1_var) / pl.col(exp2_var)).alias(exp1_var + '/' + exp2_var))

    return result
    # return result.drop([exp1_var, exp2_var])


def perform_operation_on_scalars(op, exp1, exp2):
    if op == Op.Plus:
        return exp1 + exp2
    elif op == Op.Minus:
        return exp1 - exp2
    elif op == Op.Multiply:
        return exp1 * exp2
    elif op == Op.Divide:
        return exp1 / exp2


from muse.scanner import Position

class CompileError:
    def __init__(self, msg: str, begin_pos: Position, is_warning=False):
        self.msg = msg
        self.begin_pos = begin_pos
        self.is_warning = is_warning

    def __str__(self):
        return '@{} : {}'.format(self.begin_pos, self.msg)


class RunError:
    def __init__(self, msg: str, begin_pos: Position, is_warning=False):
        self.msg = msg
        self.begin_pos = begin_pos
        self.is_warning = is_warning

    def __str__(self):
        return '@{} : {}'.format(self.begin_pos, self.msg)


if __name__ == '__main__':
    a = [1, 2]
    b = [3, 4]
    print(a + b)



demos = list()

demo0 = '''
开始日期 = '2024-04-01'
结束日期 = '2024-04-01'
公募产品开放产品 = 产品(['公募开放式'], ['净资产', '前一日净资产'])
公募产品净资产 = 公募产品开放产品.净资产
'''

# 资产指标计算
demo1 = '''回购资产 = 资产(['质押式回购', '买断式回购'], ['剩余期限', '久期']) in 产品([], ['前一日净资产'])
回购资产最大存续期 = agg(回购资产.剩余期限, ['产品代码', '产品简称'], 'max')
监控规则 = {'不符合监管要求': '剩余期限 > 365', '正常': '剩余期限 <= 365'}
'''
demos.append({'label': '1.质押式回购与买断式回购（含协议式回购）最长期限均不得超过365天', 'value': demo1})

# 产品指标计算
demo2 = '''正回购 = 资产(['债券正回购']) in 产品(['公募开放式'])
正回购持仓市值 = 正回购.持仓市值
正回购持仓市值合计 = agg(正回购持仓市值, ['产品代码', '产品简称'], 'sum')
公募产品开放产品净资产 = 产品(['公募开放式']).净资产
正回购余额占比 = 正回购持仓市值合计 / 公募产品开放产品净资产
监控规则 = {'不符合监管要求’': '正回购余额占比 > 0.4',  '正常': '正回购余额占比 <= 0.4'}
'''
demos.append({'label': '2.公募（除封闭式外）理财产品债券正回购资金余额不超过产品净资产的40%', 'value': demo2})

# 主体指标计算
demo3 = '''债券逆回购 = 资产(['债券逆回购']) in 产品(['私募'])
逆回购持仓市值 = 债券逆回购.持仓市值
逆回购持仓市值合计 = agg(逆回购持仓市值, ['产品代码', '产品简称'], 'sum')
私募产品净资产 = 产品(['私募']).净资产
逆回购余额占比 = 逆回购持仓市值合计 / 私募产品净资产
监控规则 = {'不符合监管要求': '逆回购余额占比 > 1',  '正常': '逆回购余额占比 <= 1'}
'''
demos.append({'label': '3.私募产品债券逆回购资金余额不超过产品净资产的100%', 'value': demo3})

# 资产-产品混合计算
demo4 = '''债权类资产持仓明细 = (资产(['债权类']) in 产品(['固定收益类'])).持仓市值
债权类资产持仓汇总 = agg(债权类资产持仓明细, ['产品代码', '产品简称'], 'sum')
固定收益类产品净资产 = 产品(['固定收益类']).净资产
债权类资产占比 = 债权类资产持仓汇总 / 固定收益类产品净资产
监控规则 = {'不符合监管要求': '债权类资产占比 < 0.8',  '正常': '债权类资产占比 >= 0.8'}
'''
demos.append({'label': '4.固定收益类理财产品投资于存款、债券等债权类资产的比例不低于80%', 'value': demo4})

# 主体-产品混合计算
demo5 = '''未上市股权 = 资产(['未上市企业股权及其受(收)益权']) in 产品(['公募开放式'])
未上市股权个数 = agg(未上市股权, ['产品代码', '产品简称'], 'count', {'资产代码': '未上市股权个数'})
监控规则 = {'不符合监管要求': '未上市股权个数 > 0',  '正常': '未上市股权个数 == 0'}
'''
demos.append({'label': '5.公募产品不得投资未上市企业股权', 'value': demo5})

# 这个有点小问题
demo6 = '''产品数据 = 产品(['封闭式']).下一开放日
非标数据 = 资产(['非标准化债权类资产']).资产到期日
非标持仓 = 非标数据 in 产品数据 
超标持仓 = filter(非标持仓, '资产到期日 > 下一开放日')
超限个数 = agg(超标持仓, ['产品代码', '产品简称'], 'count', {'资产代码': '超限个数'})
监控规则 = {'不符合监管要求': '超限个数 > 1',  '正常': '超限个数 <= 1'}
'''
demos.append({'label': '6.非标准化债权类资产终止日不得晚于封闭式资产管理产品的到期日', 'value': demo6})

# 合规限额计算2
demo7 = '''全部资产投资金额 = agg((资产([], '穿透') in 产品([])).持仓市值, ['资产代码', '资产简称'], 'sum')
监控规则 = {'不符合监管要求': '持仓市值 > 30000000000',  '正常': '持仓市值 <= 30000000000'}
'''
demos.append({'label': '7.【穿透】所有资产管理产品投资同一资产的资金总规模合计不得超过300亿元', 'value': demo7})

# 压力测试计算
demo8 = '''单个证券资产持仓占比 = (资产(['债券', '股票', '基金']) in 产品()).持仓市值 / 产品().净资产
监控规则 = {'不符合监管要求': '单个证券资产持仓占比 > 0.1',  '正常': '单个证券资产持仓占比 <= 0.1'}
'''
demos.append({'label': '8.单只公募理财产品持有单只证券或单只证券投资基金的市值不得超过该理财产品净资产的10%', 'value': demo8})

# 绩效评价计算
demo9 = '''证券持仓明细 = (资产(['债券', '股票', '基金']) in 产品()).持仓市值
证券持仓汇总 = agg(证券持仓明细, ['资产代码', '资产简称'], 'sum')
证券发行市值 = 资产(['债券', '股票', '基金']).发行市值
持仓占发行市值比 = 证券持仓汇总 / 证券发行市值
监控规则 = {'不符合监管要求': '持仓占发行市值比 > 0.3',  '正常': '持仓占发行市值比 <= 0.3'}
'''
demos.append({'label': '9.全部公募理财产品持有单只证券或单只证券投资基金的市值，不得超过该证券市值或该证券投资基金市值的30%', 'value': demo9})

demo10 = '''股票持仓明细 = (资产(['股票']) in 产品(['公募开放式'])).持仓市值
股票持仓汇总 = agg(股票持仓明细, ['资产代码', '资产简称'], 'sum')
股票流通市值 = 资产(['股票']).可流通市值
持仓占流通市值比 = 股票持仓汇总 / 股票流通市值
监控规则 = {'不符合监管要求': '持仓占流通市值比 > 0.15',  '正常': '持仓占流通市值比 <= 0.15'}
'''
demos.append({'label': '10.全部开放式公募理财产品投资单一上市公司发行的股票不得超过该上市公司可流通股票的15%', 'value': demo10})

# 错误示例1
demo11 = '''包括非机构投资者的产品 = filter(产品().个人投资者占比, '个人投资者占比 > 0')
不良资产 = 资产(['不良资产受(收)益权'])
不良资产持仓 = 不良资产 in 包括非机构投资者的产品
不良资产个数 = agg(不良资产持仓, ['产品代码', '产品简称'], 'count')
监控规则 = {'不符合监管要求': '个人投资者占比 > 0',  '正常': '个人投资者占比 == 0'}
'''
demos.append({'label': '11.面向非机构投资者发行的理财产品不得直接或间接投资于不良资产受（收）益权', 'value': demo11})

# 错误示例2
demo12 = '''债券资产持仓 = (资产(['债券', '资产支持证券']) in 主体() in 产品(['现金管理类'])).持仓市值
主体持仓汇总 = agg(债券资产持仓, ['产品代码', '产品简称', '主体代码', '主体名称'], 'sum')
产品净资产 = 产品(['现金管理类']).净资产
主体持仓汇总占比 = 主体持仓汇总 / 产品净资产
监控规则 = {'不符合监管要求': '主体持仓汇总占比 > 0.1',  '正常': '主体持仓汇总占比 <= 0.1'}
'''
demos.append({'label': '12.每只现金管理类产品投资于同一机构发行的债券及其作为原始权益人的资产支持证券的比例合计不得超过该产品资产净值的10%', 'value': demo12})

demo13 = '''主体AAA以下持仓 = (资产() in 主体(['主体信用评级低于AAA']) in 产品(['现金管理类'])).持仓市值
主体持仓汇总 = agg(主体AAA以下持仓, ['产品代码', '产品简称', '主体代码', '主体名称'], 'sum')
产品净资产 = 产品(['现金管理类']).净资产
主体持仓汇总占比 = 主体持仓汇总 / 产品净资产
监控规则 = {'不符合监管要求': '主体持仓汇总占比 > 0.02',  '正常': '主体持仓汇总占比 <= 0.02'}
'''
demos.append({'label': '13.每只现金管理类产品投资于所有主体信用评级低于AAA的机构发行的金融工具的比例合计不得超过该产品资产净值的10%', 'value': demo13})

demo14 = '''存款存单持仓合计 = agg((资产(['银行存款', '同业存单']) in 主体() in 产品(['现金管理类'])).持仓市值, ['主体代码', '主体名称'], 'sum')
银行上季度净资产 = agg((资产(['银行存款', '同业存单']) in 主体() in 产品(['现金管理类'])).上一季度末净资产, ['主体代码', '主体名称'], 'first')
持仓占净资产比例 = 存款存单持仓合计 / 银行上季度净资产
监控规则 = {'不符合监管要求': '持仓占净资产比例 > 0.1',  '正常': '持仓占净资产比例 <= 0.1'}
'''
demos.append({'label': '14.全部现金管理类产品投资于同一商业银行的存款、同业存单和债券，不得超过该商业银行最近一个季度末净资产的10%', 'value': demo14})

demo15 = '''高流动性资产持仓 = (资产(['现金', '国债', '中央银行票据', '政策性金融债券', '5个交易日内到期'], 去重=True) in 产品(['现金管理类'])).持仓市值
持仓合计 = agg(高流动性资产持仓, ['产品代码', '产品简称'], 'sum')
高流动性资产持仓比例 = 持仓合计 / 产品(['现金管理类']).净资产
监控规则 = {'不符合监管要求': '高流动性资产持仓比例 < 0.1',  '正常': '高流动性资产持仓比例 >= 0.1'}
'''
demos.append({'label': '15.每只现金管理类产品持有不低于该产品资产净值10%的现金、国债、中央银行票据、政策性金融债券以及5个交易日内到期的其他金融工具', 'value': demo15})

demo16 = '''筛选后产品 = filter(产品(['现金管理类']).前10名投资者份额占比, '前10名投资者份额占比 > 20')
全部资产 = 资产()
全部持仓资产剩余期限 = (全部资产 in 筛选后产品).剩余期限
剩余期限平均值 = agg(全部持仓资产剩余期限, ['产品代码', '产品简称'], 'mean')
监控规则 = {'不符合监管要求': '剩余期限 > 180',  '正常': '剩余期限 <= 180'}
'''
demos.append({'label': '16.当前10名投资者的持有份额合计超过现金管理类产品总份额的20%时，平均剩余存续期限不得超过180天', 'value': demo16})

demo17 = '''定开产品 = filter(产品(['定期开放式']).产品开放期, '产品开放期 < 90')
不活跃资产 = 资产(['不存在活跃交易市场资产'])
不活跃资产持仓 = (不活跃资产 in 定开产品).持仓市值
不活跃资产合计持仓 = agg(不活跃资产持仓, ['产品代码', '产品简称'], 'sum')
定开产品净资产 = 定开产品.净资产
不活跃资产占比 = 不活跃资产合计持仓 / 定开产品净资产
监控规则 = {'不符合监管要求': '不活跃资产占比 > 0.5',  '正常': '不活跃资产占比 <= 0.5'}
'''
demos.append({'label': '17.定期开放周期小于90天的开放式理财产品，投资不存在活跃交易市场资产的比例不得超过该理财产品净资产的50%', 'value': demo17})

demo18 = '''摊余成本法资产 = filter((资产([], ['估值方法', '资产到期日']) in 产品(['封闭式'])), '估值方法=="摊余成本法"')
封闭式持仓资产到期日 = 摊余成本法资产.资产到期日
封闭式产品到期日 = 产品(['封闭式']).产品到期日
天数差 = 封闭式产品到期日 - 封闭式持仓资产到期日
监控规则 = {'不符合监管要求': '天数差 < 0',  '正常': '天数差 >= 0'}
'''
demos.append({'label': '18.封闭式理财产品所投资的以摊余成本法计量的金融资产到期日不得晚于理财产品到期日', 'value': demo18})

demo19 = '''封闭式持仓资产到期日 = (资产() in 产品(['封闭式'])).资产到期日
城投企业持仓 = (资产() in 主体(['城投企业']) in 产品()).持仓市值
城投企业持仓合计 = sum(城投企业持仓)
产品总量 = sum(产品().净资产)
城投企业集中度 = 城投企业持仓合计 / 产品总量
房地产企业持仓 = (资产() in 主体(['房地产企业']) in 产品()).持仓市值
房地产企业持仓合计 = sum(房地产企业持仓)
房地产企业集中度 = 房地产企业持仓合计 / 产品总量
'''
demos.append({'label': '19.理财业务单一城投企业投资集中度', 'value': demo19})

demo20 = '''资产代码 = 'B7'
产品代码 = 'P07'
摊余成本法资产 = filter((资产() in 产品(['封闭式'])).估值方法, '估值方法=="摊余成本法"')
封闭式持仓资产到期日 = 摊余成本法资产.资产到期日
封闭式产品到期日 = 产品(['封闭式']).产品到期日
天数差 = 封闭式产品到期日 - 封闭式持仓资产到期日
监控规则 = {'不符合监管要求': '天数差 < 0',  '正常': '天数差 >= 0'}
'''
demos.append({'label': '20.将产品代码和资产代码录入进行溯源演示', 'value': demo20})

demo21 = '''产品代码 = '2101AF0018'
封闭式产品 = 产品(['封闭式产品'], ['当日份额', '当日单位净值'])
'''

demo22 = '''债权资产 = 资产(['单一资产'], ['持仓市值', '面值', '数量'])
'''

demo23 = '''持仓 = 资产(['单一资产'], ['持仓市值', '面值', '数量']) in 产品(['封闭式产品'], ['当日份额', '当日单位净值'])
'''

demo24 = '''查询资产 = 资产(['单一资产'], ['持仓市值'])
查询产品 = 产品(['封闭式产品'], ['当日份额', '当日单位净值'])
持仓 = 查询资产 in 查询产品
汇总 = agg(持仓, ['产品代码'], {'持仓市值': 'sum', '当日份额': 'first'})
'''
def get_demos():
    return demos



# 指标提取引擎
# 根据金融对象.指标的方法提取金融对象的数值类指标
# 同一类金融对象需要一次性提取避免性能的额外开销
import logging
import polars as pl

def indicator_handle(base, prop):
    if isinstance(base, pl.DataFrame) and prop not in base.columns:
        logging.error(f'调用点运算失败：{prop} 不在 {base}中')
        return prop, pl.DataFrame()
    return prop, base



# 语法分析器：根据token串生成AST树(树的每个节点是一个AST Node)
from muse.scanner import TokenKind, Op, Seperator, Scanner, Position
from muse.ast_struct import (Prog, ErrorStatement, ExpressionStatement, Binary, Unary, Variable, ArrayLiteral, ErrorExp, DotExp,
                        IntegerLiteral, DecimalLiteral, StringLiteral, FunctionCall, MapLiteral, EmptyStatement)
from muse.error import CompileError


class Parser:
    def __init__(self, scanner: Scanner):
        self.scanner = scanner
        self.errors = list()
        self.warnings = list()

    def add_error(self, msg: str, pos: Position):
        self.errors.append(CompileError(msg, pos, False))
        print('@{} : {}'.format(pos, msg))

    def add_warning(self, msg: str, pos: Position):
        self.warnings.append(CompileError(msg, pos, True))
        print('@{} : {}'.format(pos, msg))

    # 解析整个程序，语法分析器主入口
    def parse_prog(self):
        begin_pos = self.scanner.peek().pos
        statements = self.parse_statement_list()
        return Prog(begin_pos, self.scanner.get_last_pos(), statements)

    def parse_statement_list(self):
        statements = []
        t = self.scanner.peek()
        while t.kind != TokenKind.EOF:
            statement = self.parse_statement()
            if isinstance(statement, ErrorStatement):
                statements = list()
                break
            else:
                statements.append(statement)
            t = self.scanner.peek()
        return statements

    def parse_statement(self):
        t = self.scanner.peek()
        if t.kind == TokenKind.Identifier or t.kind == TokenKind.DecimalLiteral or t.kind == TokenKind.IntegerLiteral or t.kind == TokenKind.StringLiteral or t.code == Seperator.OpenParen:
            return self.parse_expression_statement()
        elif t.code == Seperator.SemiColon:
            self.scanner.next()
            return EmptyStatement(self.scanner.get_last_pos())
        else:
            txt = self.scanner.peek().text.strip()
            txt = '空字符' if txt == '' else txt
            self.add_error('无法识别语句, 语句开始于 ' + txt, self.scanner.get_last_pos())
            begin_pos = self.scanner.get_next_pos()
            self.skip()
            return ErrorStatement(begin_pos, self.scanner.get_last_pos())

    def parse_expression_statement(self):
        exp = self.parse_expression()
        t = self.scanner.peek()
        statement = ExpressionStatement(self.scanner.get_last_pos(), exp)
        # 看看在真实的语言中可不可以把这个去掉，不用;作为结束
        if t.code == Seperator.SemiColon:
            self.scanner.next()
        else:
            txt = t.text.strip()
            txt = '空字符' if txt == '' else txt
            self.add_error('非法语句结束符号 ' + txt, self.scanner.get_last_pos())
            self.skip()
            statement.end_pos = self.scanner.get_last_pos()
            statement.is_error = True
        return statement

    # 重点来了解析表达式
    def parse_expression(self):
        return self.parse_assignment()

    # 二元运算符的优先级
    op_priority = {
        Op.Assign: 2,
        Op.Plus: 12,
        Op.Minus: 12,
        Op.Divide: 13,
        Op.Multiply: 13,
        Op.In: 4
    }

    def get_priority(self, op):
        if op in self.op_priority:
            return self.op_priority[op]
        else:
            return -1

    # 解析赋值表达式 -- 这段代码我看不太懂
    def parse_assignment(self):
        assign_priority = self.get_priority(Op.Assign)
        # 先解析一个优先级更高的表达式
        exp1 = self.parse_binary(assign_priority)
        t = self.scanner.peek()
        t_priority = self.get_priority(t.code)
        # 存放赋值运算符量表的表达式
        exp_stack = list()
        exp_stack.append(exp1)
        # 存放赋值运算符
        op_stack = list()

        # 解析赋值表达式
        while t.kind == TokenKind.Operator and t_priority == assign_priority:
            op_stack.append(t.code)
            self.scanner.next()  # 跳过运算符
            # 获取运算符优先级高于赋值的二元表达式
            exp1 = self.parse_binary(assign_priority)
            exp_stack.append(exp1)
            t = self.scanner.peek()
            t_priority = self.get_priority(t.code)

        # 组装右结合的AST
        exp1 = exp_stack[-1]
        if len(op_stack) > 0:
            i = len(exp_stack) - 2
            while i >= 0:
                exp1 = Binary(op_stack[i], exp_stack[i], exp1)
                i = i - 1
        return exp1

    # 采用运算符优先级算法，解析二元表达式。
    # 这是一个递归算法。一开始，提供的参数是最低优先级，
    def parse_binary(self, prec: int):
        exp1 = self.parse_unary()
        t = self.scanner.peek()
        tprec = self.get_priority(t.code)

        while t.kind == TokenKind.Operator and tprec > prec:
            self.scanner.next()  # 跳过运算符
            exp2 = self.parse_binary(tprec)
            exp = Binary(t.code, exp1, exp2)
            exp1 = exp
            t = self.scanner.peek()
            tprec = self.get_priority(t.code)

        return exp1

    # 解析一元运算
    def parse_unary(self):
        begin_pos = self.scanner.get_next_pos()
        t = self.scanner.peek()
        if t.code == Op.Minus:
            self.scanner.next()
            exp = self.parse_unary()
            return Unary(begin_pos, self.scanner.get_last_pos(), t.code, exp, is_prefix=True)
        else:
            exp = self.parse_primary()
            return exp

    # 解析基础表达式
    def parse_primary(self):
        begin_pos = self.scanner.get_next_pos()
        t = self.scanner.peek()
        exp = None

        if t.kind == TokenKind.IntegerLiteral or t.kind == TokenKind.DecimalLiteral or t.kind == TokenKind.StringLiteral:
            exp = self.parse_literal()
        elif t.kind == TokenKind.Identifier:
            if self.scanner.peek2().code == Seperator.OpenParen:
                exp = self.parse_function_call()
            else:
                self.scanner.next()
                exp = Variable(begin_pos, self.scanner.get_last_pos(), t.text)
        elif t.code == Seperator.OpenParen:  # (
            self.scanner.next()
            exp = self.parse_expression()
            t1 = self.scanner.peek()
            if t1.code == Seperator.CloseParen:  # )
                self.scanner.next()
            else:
                txt = t.text.strip()
                txt = '空字符' if txt == '' else txt
                self.add_error('在主表达式的末尾需要一个), 但是程序找到了: ' + txt, self.scanner.get_last_pos())
                self.skip()
        elif t.code == Seperator.OpenBracket:  # [ 解析数组
            self.scanner.next()
            t1 = self.scanner.peek()
            exps = list()
            while t1.code != Seperator.CloseBracket and t1.kind != TokenKind.EOF:  # 添加EOF检查
                exp1 = self.parse_expression()
                exps.append(exp1)

                t1 = self.scanner.peek()
                if t1.code == Op.Comma:
                    self.scanner.next()
                    t1 = self.scanner.peek()
                elif t1.code != Seperator.CloseBracket:
                    txt = t.text.strip()
                    txt = '空字符' if txt == '' else txt
                    self.add_error('在数组末尾需要], 但是程序得到了: ' + txt, self.scanner.get_last_pos())
                    self.skip()
                    break  # 添加break以避免无限循环
            if t1.code == Seperator.CloseBracket:  # 跳过最后一个]
                self.scanner.next()
            exp = ArrayLiteral(begin_pos, self.scanner.get_last_pos(), exps)
        elif t.code == Seperator.OpenBrace:  # { 解析字典
            self.scanner.next()
            t1 = self.scanner.peek()
            exps = list()
            while t1.code != Seperator.CloseBrace and t1.kind != TokenKind.EOF:  # 添加EOF检查
                exp1 = self.parse_expression()
                k = exp1
                t1 = self.scanner.peek()
                if t1.code == Seperator.Colon:
                    self.scanner.next()
                    v = self.parse_expression()
                else:  # 如果不是:, 那肯定出错了
                    txt = t.text.strip()
                    txt = '空字符' if txt == '' else txt
                    self.add_error('字典的其中一个元素要包含:把key和value分开, 但是程序得到了: ' + txt,
                                   self.scanner.get_last_pos())
                    self.skip()
                    v = None
                    break  # 添加break以避免无限循环
                exps.append({'key': k, 'val': v})
                t1 = self.scanner.peek()
                if t1.code == Op.Comma:
                    self.scanner.next()
                    t1 = self.scanner.peek()
                elif t1.code != Seperator.CloseBrace:
                    txt = t.text.strip()
                    txt = '空字符' if txt == '' else txt
                    self.add_error('在字典末尾需要}, 但是程序得到了: ' + txt, self.scanner.get_last_pos())
                    self.skip()
                    break  # 添加break以避免无限循环
            if t1.code == Seperator.CloseBrace:  # 跳过最后一个}
                self.scanner.next()
            exp = MapLiteral(begin_pos, self.scanner.get_last_pos(), exps)
        else:
            # 遇到一些负载First集合中的Token
            txt = t.text.strip()
            txt = '空字符' if txt == '' else txt
            self.add_error('无法识别语法: ' + txt, self.scanner.get_last_pos())
            exp = ErrorExp(begin_pos, self.scanner.get_last_pos())

        # 带有后缀的表达式，比如数组
        t = self.scanner.peek()
        while t.code == Op.Dot and t.kind != TokenKind.EOF:  # 添加EOF检查
            self.scanner.next()
            exp1 = self.parse_primary()
            exp = DotExp(begin_pos, self.scanner.get_last_pos(), exp, exp1)
            t = self.scanner.peek()

        return exp

    def skip(self):
        t = self.scanner.peek()
        while t.kind != TokenKind.EOF:
            if t.kind == TokenKind.Keyword:
                return
            elif t.kind == TokenKind.Seperator and (
                    t.text == ',' or t.text == ';' or t.text == '{' or t.text == '}' or t.text == '(' or t.text == ')'):
                return
            else:
                self.scanner.next()
                t = self.scanner.peek()

    # 解析字面量
    def parse_literal(self):
        begin_pos = self.scanner.get_next_pos()
        t = self.scanner.peek()
        # 知识点：以Identifier开头，可能是函数调用，也可能是一个变量，所以要再多向后看一个Token，
        # 这相当于在局部使用了LL(2)算法。
        if t.kind == TokenKind.IntegerLiteral:
            self.scanner.next()
            return IntegerLiteral(begin_pos, int(t.text))
        elif t.kind == TokenKind.DecimalLiteral:
            self.scanner.next()
            return DecimalLiteral(begin_pos, float(t.text))
        else:
            self.scanner.next()
            return StringLiteral(begin_pos, t.text)

    # 解析函数调用
    def parse_function_call(self):
        begin_pos = self.scanner.get_next_pos()
        params = list()
        name = self.scanner.next().text

        self.scanner.next()  # 跳过(
        # 循环读出所有参数
        t1 = self.scanner.peek()
        while t1.code != Seperator.CloseParen and t1.kind != TokenKind.EOF:
            exp = self.parse_expression()
            params.append(exp)

            if exp.is_error:
                self.add_error('在解析函数调用时出现问题: ' + name, self.scanner.get_last_pos())
            t1 = self.scanner.peek()
            if t1.code != Seperator.CloseParen:
                if t1.code == Op.Comma:
                    t1 = self.scanner.next()
                else:
                    txt = t1.text.strip()
                    txt = '空字符' if txt == '' else txt
                    self.add_error('函数调用末尾需要一个), 但程序得到了一个 ' + txt, self.scanner.get_last_pos())
                    self.skip()
                    return FunctionCall(begin_pos, self.scanner.get_last_pos(), name, params, True)

        if t1.code == Seperator.CloseParen:
            self.scanner.next()

        return FunctionCall(begin_pos, self.scanner.get_last_pos(), name, params)

    def skip(self):
        t = self.scanner.peek();
        while t.kind != TokenKind.EOF:
            if t.kind == TokenKind.Keyword:
                return
            elif t.kind == TokenKind.Seperator and (t.text == ',' or t.text == ';' or t.text == '{' or t.text == '}' or t.text == '(' or t.text == ')'):
                return
            else:
                self.scanner.next()
                t = self.scanner.peek()






# 运行解释器：遍历AST节点，根据树结构顺序分别解析运行
# 调用tag_engine(标签处理), ind_engine(指标获取), calc_engine(四则运算引擎), algo_engine(金融算法引擎)来对语句进行处理
import traceback
from muse.ast_struct import AstVisitor,  Binary, Unary, Variable, FunctionCall, DotExp
from muse.scanner import Op
from muse.engines import tag_engine, algo_engine, calc_engine, indicator_engine
from collections import OrderedDict
import logging
from datetime import datetime


blanks = ' ' * 4


# 打印AST的调试程序
class Interpreter(AstVisitor):
    def __init__(self, props):
        super().__init__()
        self.results = OrderedDict()
        self.props = set(props)
        self.start_date = datetime.now().date().strftime('%Y-%m-%d')
        self.end_date = self.start_date
        self.search_mode = '单元测试'

    def visit_function_call(self, func_call: FunctionCall, additional):
        # 获得函数调用名称
        name = func_call.name
        ps = list()
        for p in func_call.params:
            ps.append(self.visit(p, additional))

        try:
            #  如果是以下三个函数调用，则使用
            tags, fields, pen, drop_duplicate = tag_engine.parse_tags_and_fields(func_call.params)
            props = list(set(fields + list(self.props)))
            if name == '资产':
                if '资产代码' in self.results:
                    return tag_engine.asset_tag(tags, props, sec_ids=self.results['资产代码'], start_date=self.start_date, end_date=self.end_date, drop_duplicate=drop_duplicate, search_mode=self.search_mode)
                else:
                    return tag_engine.asset_tag(tags, props, start_date=self.start_date, end_date=self.end_date, drop_duplicate=drop_duplicate, search_mode=self.search_mode)
            elif name == '主体':
                return tag_engine.issuer_tag(tags, props, start_date=self.start_date, end_date=self.end_date, drop_duplicate=drop_duplicate, search_mode=self.search_mode)
            elif name == '产品':
                if '产品代码' in self.results:
                    return tag_engine.port_tag(tags, props, port_ids=self.results['产品代码'], start_date=self.start_date, end_date=self.end_date, drop_duplicate=drop_duplicate, search_mode=self.search_mode)
                return tag_engine.port_tag(tags, props, start_date=self.start_date, end_date=self.end_date, drop_duplicate=drop_duplicate, search_mode=self.search_mode)
            elif name == 'filter':
                return tag_engine.filter_algo(ps, self.props)
            else:
                return algo_engine.algo_handler(name, ps)
        except Exception:
            self.add_error(traceback.format_exc(), func_call.begin_pos)
            return

    def visit_binary(self, binary: Binary, additional):
        exp1 = binary.exp1
        exp2 = binary.exp2
        op = binary.op
        sec_ids = None
        port_ids = None
        if '资产代码' in self.results:
            sec_ids = self.results['资产代码']
        if '产品代码' in self.results:
            port_ids = self.results['产品代码']
        # 如果是in操作，要保证左侧和右侧都是函数调用
        if op == Op.In and isinstance(exp1, FunctionCall) and isinstance(exp2, FunctionCall):
            try:
                result = tag_engine.in_tag(exp1, exp2, self.props, sec_ids=sec_ids, port_ids=port_ids, start_date=self.start_date, end_date=self.end_date, search_mode=self.search_mode)
            except Exception:
                self.add_error(traceback.format_exc(), binary.begin_pos)
                result = None
        elif op == Op.In and isinstance(exp1, Binary) and isinstance(exp2, FunctionCall):
            try:
                result = tag_engine.in_tag2(exp1, exp2, self.props, sec_ids=sec_ids, port_ids=port_ids, start_date=self.start_date, end_date=self.end_date, search_mode=self.search_mode)
            except Exception:
                self.add_error(traceback.format_exc(), binary.begin_pos)
                result = None
        elif op == Op.In and isinstance(exp1, Variable) and isinstance(exp2, Variable):
            try:
                result = tag_engine.in_tag3(self.results[exp1.name], self.results[exp2.name], self.props, sec_ids=sec_ids, port_ids=port_ids, start_date=self.start_date, end_date=self.end_date, search_mode=self.search_mode)
            except Exception:
                self.add_error(traceback.format_exc(), binary.begin_pos)
                result = None
        else:
            v1 = self.visit(exp1, None)
            v2 = self.visit(exp2, None)
            if op == Op.Assign:
                v1 = self.visit(exp1, 'left')
                if v1 == '开始日期':
                    self.start_date = v2
                    logging.debug('调用赋值表达式, 把 {} 放入 {}'.format(v2, 'start_date'))
                elif v1 == '结束日期':
                    self.end_date = v2
                    logging.debug('调用赋值表达式, 把 {} 放入 {}'.format(v2, 'end_date'))
                elif v1 == '查询模式':
                    self.search_mode = v2
                else:
                    self.results[v1] = v2
                    logging.debug('调用赋值表达式, 把 {} 放入 {}'.format(v1, self.results))

                result = v2
            else:
                try:
                    result = calc_engine.handle_binary_op(op, v1, v2)
                except Exception:
                    self.add_error(traceback.format_exc(), binary.begin_pos)
                    result = None
        return result

    def visit_unary(self, unary: Unary, additional):
        v = self.visit(unary.exp, additional)
        if unary.op == Op.Minus:
            return -1 * v
        else:
            self.add_error('一元运算只支持 "-"', unary.begin_pos)
            return None

    def visit_dot_exp(self, dot_exp: DotExp, additional):
        base_exp = self.visit(dot_exp.base_exp, additional)
        prop_exp = self.visit(dot_exp.prop, additional)
        try:
            return indicator_engine.indicator_handle(base_exp, prop_exp)
        except Exception:
            self.add_error(traceback.format_exc(), dot_exp.begin_pos)
            return None

    def visit_variable(self, variable: Variable, additional):
        var_name = variable.name
        if additional is not None and additional == 'left':
            return var_name
        else:
            if var_name in self.results:
                logging.debug('{} 已经被解析过直接使用. 之前解析的结果为 {}'.format(var_name, self.results[var_name]))
                return self.results[var_name]
            else:
                logging.debug('没有在{} 中 发现 {}.这一定是个需要的指标'.format(self.results, var_name))
                return var_name






import polars as pl
import pandas as pd
import numpy as np
import muse.data_interface as di
import random
from datetime import timedelta, datetime
from muse import muse_config as s

DATE = '2024-11-30'
DDATE = datetime.strptime(DATE, '%Y-%m-%d')


# 生成产品信息
def gen_port_info():
    # 生成产品信息
    port = pl.read_excel(di.TEST_LOCATE + '产品信息.xlsx')
    port_tags = list(port.select('产品标签').unique().to_series())
    ports = list()
    for i in range(0, 10000):
        p = dict()
        p['产品代码'] = 'P' + str(i)
        p['产品简称'] = '产品' + str(i)
        p['产品标签'] = random.sample(port_tags, 1)[0]
        ports.append(p)
    port_info = pl.DataFrame(ports)
    port_info.write_parquet(di.TEST_LOCATE + '产品信息.parquet')


# 生成主体信息
def gen_issuer_info():
    issuer = pl.read_excel(di.TEST_LOCATE + '主体信息.xlsx')
    tags = list(issuer.select('主体标签').unique().to_series())
    issuers = list()
    for i in range(0, 300):
        p = dict()
        p['主体代码'] = 'ISSUER' + str(i)
        p['主体标签'] = random.sample(tags, 1)[0]
        p['主体名称'] = '主体' + str(i)
        issuers.append(p)
    issuer_info = pl.DataFrame(issuers)
    issuer_info.write_parquet(di.TEST_LOCATE + '主体信息.parquet')
    print(issuer_info)


# 生成资产信息
def gen_sec_info():
    sec_info = pl.read_excel(di.TEST_LOCATE + '资产信息.xlsx')
    sec_tags = list(sec_info.select('资产标签').unique().to_series())
    issuer = pl.read_parquet(di.TEST_LOCATE + '主体信息.parquet')
    issuer_ids = list(issuer.select('主体代码').unique().to_series())
    secs = list()
    for i in range(0, 1000):
        s = dict()
        s['资产代码'] = 'SEC' + str(i)
        s['资产简称'] = '资产' + str(i)
        s['主体代码'] = random.sample(issuer_ids, 1)[0]
        s['资产标签'] = random.sample(sec_tags, 1)[0]
        secs.append(s)
    sec_info = pl.DataFrame(secs)
    sec_info.write_parquet(di.TEST_LOCATE + '资产信息.parquet')
    print(sec_info)


# 生成持仓指标
def gen_hld_ind():
    hlds = list()
    port = pl.read_parquet(di.TEST_LOCATE + '产品信息.parquet')
    port_ids = list(port.select('产品代码').unique().to_series())
    sec = pl.read_parquet(di.TEST_LOCATE + '资产信息.parquet')
    sec_ids = list(sec.select('资产代码').unique().to_series())

    for port_id in port_ids:
        selected_sec_ids = random.sample(sec_ids, len(sec_ids) - 200)
        for s_id in selected_sec_ids:
            row = dict()
            row['产品代码'] = port_id
            row['资产代码'] = s_id
            row['日期'] = DATE
            row['估值方法'] = random.sample(['摊余成本法', '市价法', '市价法', '市价法', '市价法', '市价法', '市价法'], 1)[0]
            row['持仓市值'] = random.randint(100000, 1000000)
            hlds.append(row)
    hld_df = pl.DataFrame(hlds)
    hld_df.write_parquet(di.TEST_LOCATE + '持仓指标.parquet')
    print(hld_df)


# 生成产品指标
def gen_port_ind():
    hld_df = pd.read_parquet(di.TEST_LOCATE + '持仓指标.parquet')
    sec_df = pd.read_parquet(di.TEST_LOCATE + '资产信息.parquet')
    hld_df = hld_df.merge(sec_df, on='资产代码', how='left')
    port_df = hld_df.groupby(['产品代码', '日期'], as_index=False).agg(总资产=('持仓市值', 'sum'))
    zhg_df = hld_df[hld_df.资产标签 == '债券正回购'].groupby(['产品代码', '日期'], as_index=False).agg(正回购市值=('持仓市值', 'sum'))
    port_df = port_df.merge(zhg_df, on=['产品代码', '日期'], how='left')
    port_df['净资产'] = port_df.总资产 - port_df.正回购市值
    port_df['杠杆率'] = np.round(port_df.总资产 / port_df.净资产, 4)
    port_df['前一日净资产'] = np.round(port_df.净资产 * 0.99, 2)
    port_df['上月末净资产'] = np.round(port_df.净资产 * 0.9, 2)
    port_df['产品开放期'] = np.random.randint(1, 120, size=(10000, 1))
    port_df['产品剩余期限'] = np.random.randint(360, 720, size=(10000, 1))
    port_df['产品剩余存续期'] = np.random.randint(0, 60, size=(10000, 1)) / 100
    port_df['下一开放日'] = port_df.apply(lambda r: (timedelta(days=r['产品开放期']) + DDATE).strftime('%Y-%m-%d'), axis=1)
    port_df['产品到期日'] = port_df.apply(lambda r: (timedelta(days=r['产品剩余存续期']) + DDATE).strftime('%Y-%m-%d'),axis=1)
    port_df['总份额'] = np.round(port_df.净资产 / 1000000, 0) * 1000000
    port_df['单位净值'] = np.round(port_df.净资产 / port_df.总份额, 3)
    port_df['前10名投资者份额占比'] = np.random.randint(40, 80, size=(10000, 1))
    port_df['前10名投资者的持有份额'] = port_df.前10名投资者份额占比 / 100 * port_df.总份额
    port_df['最大单一投资者份额占比'] = np.random.randint(5, 30, size=(10000, 1))
    port_df['最大单一投资者持有份额'] = port_df.最大单一投资者份额占比 / 100 * port_df.总份额
    port_df['个人投资者占比'] = np.random.randint(0, 60, size=(10000, 1)) / 100
    port_df['机构投资者占比'] = 1 - port_df.个人投资者占比
    port_df.to_parquet(di.TEST_LOCATE + '产品指标.parquet')
    print(port_df.iloc[0])


# 生成资产指标和主体指标
def gen_sec_ind():
    sec_df = pd.read_parquet(di.TEST_LOCATE + '资产信息.parquet')
    secs = sec_df[['资产代码', '主体代码']].drop_duplicates()
    issuer_df = pd.read_parquet(di.TEST_LOCATE + '主体信息.parquet')
    issuer_df = issuer_df[['主体代码', '主体名称']]
    issuer_df = issuer_df.rename(columns={'主体名称': '主体'})
    issuer_df = issuer_df.drop_duplicates()
    secs = secs.merge(issuer_df, on='主体代码', how='left')
    secs['发行人'] = secs.主体
    secs['债权人'] = secs.主体
    secs['实际信用承担方'] = secs.主体
    secs['原始权益人'] = secs.主体
    secs['剩余期限'] = np.random.randint(1, 720, size=(len(secs), 1))
    secs['存续期限'] = secs.剩余期限 + 100
    secs['资产到期日'] = secs.apply(lambda r: (timedelta(days=r['剩余期限']) + DDATE).strftime('%Y-%m-%d'), axis=1)
    secs['资产起始日'] = secs.apply(
        lambda r: (datetime.strptime(r['资产到期日'], '%Y-%m-%d') - timedelta(days=r['存续期限'])).strftime('%Y-%m-%d'),axis=1)
    hld_df = pd.read_parquet(di.TEST_LOCATE + '持仓指标.parquet')
    hld_vals = hld_df.groupby(['资产代码'], as_index=False).agg(发行市值=('持仓市值', 'sum'))
    hld_vals['发行市值'] = hld_vals.发行市值 * 5
    secs = secs.merge(hld_vals, on='资产代码', how='left')
    secs['可流通市值'] = secs['发行市值']
    secs['久期'] = np.random.randint(1, 10, size=(len(secs), 1))
    secs['日期'] = DATE
    # 生成主体数据
    issuer_df['上一季度末净资产'] = np.random.randint(20000000, 30000000, size=(len(issuer_df), 1))
    issuer_df = issuer_df[['主体代码', '上一季度末净资产']]
    issuer_df['日期'] = DATE

    secs.to_parquet(di.TEST_LOCATE + '资产指标.parquet')
    issuer_df.to_parquet(di.TEST_LOCATE + '主体指标.parquet')
    print(secs)
    print(issuer_df)


def gen_yields():
    key_points = [0.1, 0.3, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5, 6, 7, 8, 9, 10, 15, 20, 25, 30]
    points = len(key_points)
    date_range = pd.date_range(start='2024-01-01', end='2024-12-01', freq='B')
    days = len(date_range)

    # 生成利率债市场行情信息
    data = np.random.randint(10, 300, [days, points])
    data = np.round(data / 10000, 4)
    df = pd.DataFrame(data=data, index=date_range, columns=key_points)
    df.index.name = '日期'
    df.to_excel(di.TEST_LOCATE + '国债收益率曲线.xlsx')

    # 生成信用债
    data1 = np.random.randint(10, 400, [days, points])
    data1 = np.round(data1 / 10000, 4)
    df1 = pd.DataFrame(data=data1, index=date_range, columns=key_points)
    df1.index.name = '日期'
    df1.to_excel(di.TEST_LOCATE + '信用债收益率曲线.xlsx')
    return 0


def gen_index():
    date_range = pd.date_range(start='2024-01-01', end='2024-12-01', freq='B')
    days = len(date_range)
    data = np.random.randint(-300, 300, [days, 1])
    data = np.round(data / 1000, 4)
    df = pd.DataFrame(data=data, index=date_range, columns=['PCT_CHANGE'])
    df.index.name = '日期'
    df.to_excel(di.TEST_LOCATE + '沪深300指数.xlsx')
    return 0


def gen_hld_ind2():
    hld_df = pd.read_excel(s.APP_LOCATE + '/muse/data/持仓指标.xlsx')
    hlds = hld_df[hld_df.产品代码.isin(['P01', 'P02'])]
    hlds = hlds.drop(columns=['估值方法'])
    hlds = hlds.rename(columns={'产品代码': 'PORT_ID', '资产代码': 'SEC_ID', '日期': 'VALID_DATE', '持仓市值': 'FULL_MV'})
    tags = np.random.choice(['利率债', '信用债', '股票'], len(hlds))
    hlds['AST_TAG'] = tags
    hlds['REMAIN_DAYS'] = np.nan
    hlds['MOD_DURATION'] = np.nan
    hlds['CONVEXITY'] = np.nan
    hlds['BETA'] = np.nan

    stock_len = len(hlds[hlds.AST_TAG == '股票'])
    bond_len = len(hlds) - stock_len
    hlds.loc[hlds.AST_TAG == '股票', ['BETA']] = np.random.randint(50, 200, [stock_len, 1]) / 100
    hlds.loc[hlds.AST_TAG != '股票', ['MOD_DURATION']] = np.random.randint(1, 300, [bond_len, 1]) / 100 + 1
    hlds.loc[hlds.AST_TAG != '股票', ['CONVEXITY']] = np.random.randint(100, 500, [bond_len, 1]) / 10 + 1
    hlds.loc[hlds.AST_TAG != '股票', ['REMAIN_DAYS']] = np.random.randint(1, 300, [bond_len, 1]) / 100 + 1
    hlds['NET_ASSET'] = hlds.groupby(['PORT_ID'])['FULL_MV'].transform('sum') + 10000000
    hlds.to_excel(di.TEST_LOCATE + '持仓数据.xlsx', index=False)
    return 0


if __name__ == '__main__':
    # gen_port_info()
    # gen_issuer_info()
    # gen_sec_info()
    # gen_hld_ind()
    # gen_port_ind()
    # gen_sec_ind()
    print(gen_yields())
    gen_hld_ind2()
    gen_index()